(* Wolfram Language Package *)

BeginPackage[
	"QuantumAlgebra`Agreements`", 
	{ 
		"QuantumAlgebra`"
	}
]

(* Exported symbols added here with SymbolName::usage *)  

(* ::Section:: *)
(* Exported Symbols *)



QASelectDefault::usage = 
	"QASelectDefault[setname] selects a default set of Assumptions, Index Character, Explicit " <>
	"Integration Variables, Time Constraint, Not Scripted Symbols. \n" <>
	"Those settings must be defined subsequently and are assigned to the current Notebook. " <>
	"The name chosen for the set must be unique for different Notebooks so that the Notebooks " <>
	"may be opened at the same time. This is also used as Context for the Notebook " <>
	"to keep certain definitions unique."

QAGetDefault::usage =
	"QAGetDefault[] returns the default set name."

QAReset::usage =
	"QAReset[] resets the settings of the current Notebook."

QAGetIndexCharacter::usage = 
	"QAGetIndexCharacter[set, default] returns the Index Character for the given set. Preferably " <>
	"called for the default set. A default setting can be specified."

QASetIndexCharacter::usage =
	"QASetIndexCharacter[characterdef, set] sets the Index Character of the given set. Preferably " <>
	"called for the default set."

QAResetReplacementRules::usage =
	"QAResetReplacementRules[set] resets the given set of Replacement Rules."

QAAddReplacementRules::usage =
	"QAAddReplacementRules[rules, set] adds to the given set of Replacement Rules."

QAGetReplacementRules::usage =
	"QAGetReplacementRules[set] gets the given set of Replacement Rules."

QAResetExplicitIntegrationVariables::usage = 
	"QAResetExplicitIntegrationVariables[set] resets the given set of Explicit Integration Variables."

QAAddExplicitIntegrationVariables::usage = 
	"QAAddExplicitIntegrationVariables[vars, set] adds vars to the given set of Explicit Integration Variables."

QAGetExplicitIntegrationVariables::usage = 
	"QAGetExplicitIntegrationVariables[set] gets the given set of Explicit Integration Variables."

QAResetAssumptions::usage = 
	"QAResetAssumptions[set] resets the given set of Assumptions."

QAAddAssumptions::usage = 
	"QAAddAssumptions[ass, set] adds a list of assumptions to the given set of Assumptions."

QAGetAssumptions::usage = 
	"QAGetAssumptions[set] gets the given set of Assumptions."

QASetTimeConstraint::usage =
	"QASetTimeConstraint[tc, set] defines the Time Constraint tc for QASimplify (default: 300.)."

QAGetTimeConstraint::usage =
	"QAGetTimeConstraint[set] gets the Time Constraint for QASimplify (default: 300.)."

QAResetNotScriptedSymbols::usage = 
	"ResetNotScriptedSymbols[set] resets the set of Symbols not to be used as scripted symbols."

QAAddNotScriptedSymbols::usage = 
	"AddNotScriptedSymbols[sym, set] adds the symbol list sym to the set of Symbols not to be used as scripted symbols."

QAGetNotScriptedSymbols::usage = 
	"GetNotScriptedSymbols[set] gets the set of Symbols not to be used as scripted symbols."


(* ::Section:: *)
(* Implementation of the Package *)

Begin["`Private`"] (* Begin Private Context *) 

		

(* ::Subsection:: *)
(* Public Rules *)

(* The pure version of set selection *)
QASelectDefault[st_String : "default"] := 
	Module[{},
		defineContext[st];
		set[default, st]; 
		QAReset[]
	]

(* If defined returns the default set, otherwise "empty" *)
QAGetDefault[] := getSet[Null]

(* If a re-run of initialisation is desired, one can invoke this function first.
   It enables a clean state. Call it AFTER QASelectDefault.
*)
QAReset[] :=
	Module[{},
		resetnb[notScriptedSymbols];
		resetnb[replacementRules];
		resetnb[variables];
		resetnb[timeConstraint];
		resetnb[assumptions];
		resetnb[indexCharacter];
	]

(* The pure version of get Index Character *)
QAGetIndexCharacter[set_ : Null, def: {_String, _List} : {"Mixed", {}}] := get[indexCharacter, set, def]

(* The pure version of set Index Character *)
QASetIndexCharacter[characterdef: {character_String, vars_List}, st_ : Null] := set[indexCharacter, characterdef, st]

(* Resets the given set of Replacement Rules *)
QAResetReplacementRules[set_ : Null] := reset[replacementRules, set]

(* Adds to the given set of Replacement Rules *)
QAAddReplacementRules[rules_List, set_ : Null] := add[replacementRules, rules // Flatten, set]

(* Gets the given set of Replacement Rules
   Replacement Rules can be predefined -> that's why the extra handling 
*)
QAGetReplacementRules[set_ : Null] := 
	Module[{},
		If[MatchQ[set, _String] && KeyExistsQ[replacementRules, set],
			Lookup[replacementRules, set],
			get[replacementRules, set]
		]
	]

(* Resets the given set of Integration Variables *)
QAResetExplicitIntegrationVariables[set_ : Null] := reset[variables, set]

(* Adds to the given set of Integration Variables *)
QAAddExplicitIntegrationVariables[vars_List, set_ : Null] := add[variables, vars, set]

(* Gets the given set of Integration Variables *)
QAGetExplicitIntegrationVariables[set_ : Null] := get[variables, set]

(* Reset the assumptions of a given set *)
QAResetAssumptions[set_ : Null] := reset[assumptions, set]

(* Adds to the assumptions of a given set *)
QAAddAssumptions[ass_List, set_ : Null] := add[assumptions, ass, set]

(* Gets the assumptions of a given set *)
QAGetAssumptions[set_ : Null] := get[assumptions, set]

(* Sets the Time Constraint of a given set *)
QASetTimeConstraint[constraint_, st_ : Null] := set[timeConstraint, constraint, st]

(* Gets the Time Constraint of a given set *)
QAGetTimeConstraint[set_ : Null, def_ : 300.] := get[timeConstraint, set, def]

(* Resets the given set of Not Scripted Symbols *)
QAResetNotScriptedSymbols[set_ : Null] := reset[notScriptedSymbols, set]

(* Adds to the given set of Not Scripted Symbols *)
QAAddNotScriptedSymbols[symbols_List, set_ : Null] := add[notScriptedSymbols, symbols, set]

(* Gets the given set of Not Scripted Symbols *)
QAGetNotScriptedSymbols[set_ : Null] := get[notScriptedSymbols, set]


(* ::Subsection:: *)
(* Internal Rules and Data *)

(* Gets the given set(entry) of the given association, also given a default values *)
SetAttributes[get, HoldFirst]
get[assoc_?AssociationQ, set_ : Null, def_ : {}] := Lookup[assoc, getSet[set], def]

(* Sets the given set(entry) of the given association to given value *)
SetAttributes[set, HoldFirst]
set[assoc_?AssociationQ, val_, st_ : Null] := AssociateTo[assoc, getSet[st] -> val]

(* Resets the given set(entry) of the given association to empty list *)
SetAttributes[reset, HoldFirst]
reset[assoc_?AssociationQ, set_ : Null] := AssociateTo[assoc, getSet[set] -> {}]

(* Resets the sets(entries) of the given association, belonging to current notebook, to empty list *)
SetAttributes[resetnb, HoldFirst]
resetnb[assoc_?AssociationQ] := 
	Module[{def, keys},
		def = getSet[Null];
		keys = Keys[KeySelect[assoc, StringStartsQ[#, def]&]];
		KeyDropFrom[assoc, keys]
	]

(* Adds to the given set(entry) of the given association the given list *)
SetAttributes[add, HoldFirst]
add[assoc_?AssociationQ, lst_List, set_ : Null] := 
	Module[{old, new},
		old = Lookup[assoc, getSet[set], {}];
		new = Join[old, lst];
		AssociateTo[assoc, getSet[set] -> new]
	]

(* Returns the set name *)
getSet[set_String] :=
	Module[{def},
		def = getSet[Null];
		If[def != defaultKey[], 
			def <> ":" <> set, 
			set
		]
	] 

(* Returns the default set name *)
getSet[Null] := 
	Module[{key, set},
		key = defaultKey[];
		set = Lookup[default, key, key];
		set
	]

(* Defines new context *)
defineContext[ctx_String] :=
	Module[{},
		$Context = ctx <> "`";
		$ContextPath = DeleteCases[$ContextPath, $Context];
		AppendTo[$ContextPath, $Context];
	]

(*	The Notebook Name - to be used as Association key
	Name requires unnecessary CPU resources 
	-> use Notebook as key
	-> NEW: USE CONTEXT 
*)
defaultKey[] := $Context (* EvaluationNotebook[] *)


(* ::Subsection:: *)
(* Variables *)

(* Named sets of Symbols which are not scripted *)
notScriptedSymbols = <| |>

(* Named sets of Replacement Rules *)
replacementRules = <| |>

(* Named sets of Explicit Integration / Summation Variables *)
variables = <| |>

(* Named Time Constraints for Simplify *)
timeConstraint = <| |>

(* Named sets of Assumptions for Simplify *)
assumptions = <| |>

(* Named Index Characters, grouped with related variables *)
indexCharacter = <| |>						(* Maybe Contravariant, Covariant, Mixed *)

(* The notebook's default sets *)
default = <| |>


(* End Private Context *)
End[]

EndPackage[]